from ._chunk import Chunk
from ._data import example_file_prefix, get_data_folder
from ._read import read_plink, read_plink1_bin
from ._read_grm import read_grm
from ._read_rel import read_rel
from ._testit import test
from ._write import write_plink1_bin

__version__ = "2.2.1"


__all__ = [
    "Chunk",
    "__version__",
    "example_file_prefix",
    "get_data_folder",
    "read_grm",
    "read_plink",
    "read_plink1_bin",
    "read_rel",
    "test",
    "write_plink1_bin",
]



from ._filetype import file_type


def read_rel(filepath, id_filepath=None):
    """
    Read PLINK realized relationship matrix files [b]_.

    It supports plain text, binary, and compressed files.
    The usual file extensions for those types are `.rel`, `.rel.bin`, and `.rel.zst`,
    respectively.

    Example
    -------
    .. doctest::

        >>> from os.path import join
        >>> from pandas_plink import read_rel
        >>> from pandas_plink import get_data_folder
        >>> filepath = join(get_data_folder(), "rel-bin", "plink2.rel.bin")
        >>> id_filepath = join(get_data_folder(), "rel-bin", "plink2.rel.id")
        >>> K = read_rel(filepath, id_filepath)
        >>> print(K)
        <xarray.DataArray (sample_0: 10, sample_1: 10)>
        array([[ 0.89,  0.23, -0.19, -0.01, -0.14,  0.29,  0.27, -0.23, -0.10,
                -0.21],
               [ 0.23,  1.08, -0.45,  0.19, -0.19,  0.17,  0.41, -0.01, -0.13,
                -0.13],
               [-0.19, -0.45,  1.18, -0.04, -0.15, -0.20, -0.31, -0.04,  0.30,
                -0.01],
               [-0.01,  0.19, -0.04,  0.90, -0.07,  0.01,  0.06, -0.19, -0.09,
                 0.17],
               [-0.14, -0.19, -0.15, -0.07,  1.18,  0.09, -0.03,  0.10,  0.22,
                 0.17],
               [ 0.29,  0.17, -0.20,  0.01,  0.09,  0.96,  0.07, -0.04, -0.09,
                -0.23],
               [ 0.27,  0.41, -0.31,  0.06, -0.03,  0.07,  0.71, -0.10, -0.09,
                -0.06],
               [-0.23, -0.01, -0.04, -0.19,  0.10, -0.04, -0.10,  1.42, -0.30,
                -0.07],
               [-0.10, -0.13,  0.30, -0.09,  0.22, -0.09, -0.09, -0.30,  0.91,
                -0.02],
               [-0.21, -0.13, -0.01,  0.17,  0.17, -0.23, -0.06, -0.07, -0.02,
                 0.91]])
        Coordinates:
          * sample_0  (sample_0) object 'HG00419' 'HG00650' ... 'NA20508' 'NA20753'
          * sample_1  (sample_1) object 'HG00419' 'HG00650' ... 'NA20508' 'NA20753'
            fid       (sample_1) object 'HG00419' 'HG00650' ... 'NA20508' 'NA20753'
            iid       (sample_1) object 'HG00419' 'HG00650' ... 'NA20508' 'NA20753'


    Parameters
    ----------
    filepath : str
        Path to the matrix file.
    id_filepath : str, optional
        Path to the file containing family and individual IDs. It defaults to ``None``,
        in which case it will try to be inferred.

    Returns
    -------
    rel : :class:`xarray.DataArray`
        Realized relationship matrix.

    References
    ----------
    .. [b] Read PLINK realized relationship matrix files.
           https://www.cog-genomics.org/plink/2.0/distance
    """

    ft = file_type(filepath)

    if ft == "zstd":
        return _read_rel_zs(filepath, id_filepath)
    elif ft == "bin":
        return _read_rel_bin(filepath, id_filepath)

    return _read_rel(filepath, id_filepath)


def _read_rel(filepath, id_filepath):
    df = _read_id_file(id_filepath, filepath)
    K = _1d_to_2d(_read_rel_file(filepath), df.shape[0])

    return _data_array(K, df)


def _read_rel_bin(filepath, id_filepath):
    from numpy import float64, fromfile

    df = _read_id_file(id_filepath, filepath)
    K = fromfile(filepath, dtype=float64)
    n = df.shape[0]
    K = K.reshape((n, n))

    return _data_array(K, df)


def _read_rel_file(filepath):
    from numpy import float64

    rows = []
    with open(filepath, "r") as f:
        for row in f:
            rows += [float64(v) for v in row.strip().split("\t")]
    return rows


def _read_rel_zs_rows(filepath, chunk_size=8 * 1000 * 1000):
    from zstandard import ZstdDecompressor

    with open(filepath, "rb") as fh:
        ctx = ZstdDecompressor()
        with ctx.stream_reader(fh) as reader:
            over = False
            chunks = []
            rows = []
            while not over:
                have_row = False
                while not have_row:
                    chunk = reader.read(chunk_size)
                    if not chunk:
                        over = True
                        break
                    if b"\n" in chunk:
                        have_row = True
                    chunks.append(chunk)
                (new_rows, semi_row) = _consume_rows(chunks)
                rows += new_rows
                chunks = [semi_row]
    return rows


def _consume_rows(chunks):
    from numpy import float64

    chunk = b"".join(chunks)
    rows = chunk.split(b"\n")
    semi_row = rows[-1]
    rows = [[float64(v) for v in r.split(b"\t")] for r in rows[:-1]]
    return (rows, semi_row)


def _read_rel_zs(filepath, id_filepath):
    df = _read_id_file(id_filepath, filepath)
    rows = _read_rel_zs_rows(filepath)
    K = _1d_to_2d([v for r in rows for v in r], df.shape[0])
    return _data_array(K, df)


def _data_array(K, df):
    from xarray import DataArray

    coords = (df.iloc[:, 1], df.iloc[:, 1])
    K = DataArray(K, dims=["sample_0", "sample_1"], coords=coords)
    K = K.assign_coords(**{"fid": ("sample_0", df.iloc[:, 0])})
    K = K.assign_coords(**{"fid": ("sample_1", df.iloc[:, 0])})

    K = K.assign_coords(**{"iid": ("sample_0", df.iloc[:, 1])})
    K = K.assign_coords(**{"iid": ("sample_1", df.iloc[:, 1])})

    return K


def _read_id_file(id_filepath, filepath):
    from pandas import read_csv

    if filepath.endswith(".gz"):
        basename = filepath[:-3]
    elif filepath.endswith(".bin"):
        basename = filepath[:-4]
    elif filepath.endswith(".zst"):
        basename = filepath[:-4]
    else:
        basename = filepath

    if id_filepath is None:
        id_filepath = basename + ".id"

    return read_csv(id_filepath, sep="\t", header=None, comment="#")


def _1d_to_2d(values, n):
    from numpy import tril, tril_indices_from, zeros

    K = zeros((n, n))
    K[tril_indices_from(K)] = values
    K = K + tril(K, -1).T
    return K



from hashlib import md5
from os.path import dirname, join, realpath
from pathlib import Path

from numpy import dtype, nan
from numpy.testing import assert_array_equal, assert_equal

from pandas_plink import read_plink1_bin, write_plink1_bin


def test_write_plink1_bin(tmp_path: Path):

    datafiles = join(dirname(realpath(__file__)), "data_files")
    file_prefix = join(datafiles, "data")
    bim = file_prefix + ".bim"
    bed = file_prefix + ".bed"
    fam = file_prefix + ".fam"

    G = read_plink1_bin(bed, bim, fam, verbose=False)
    assert_equal(G.data.dtype, dtype("float32"))

    snp = G.where((G.chrom == "1") & (G.pos == 72515), drop=True)["snp"].values
    assert_array_equal(snp, ["rs4030300"])

    shape = G.where(G.chrom == "1", drop=True).shape
    assert_array_equal(shape, [3, 10])

    shape = G.where(G.chrom == "2", drop=True).shape
    assert_array_equal(shape, [3, 0])

    g = G.where((G.fid == "Sample_2") & (G.iid == "Sample_2"), drop=True)
    assert_array_equal(g["trait"].values, ["-9"])

    arr = [
        [2.0, 2.0, nan, nan, 2.0, 2.0, 2.0, 2.0, 1.0, 2.0],
        [2.0, 1.0, nan, nan, 2.0, 2.0, 1.0, 2.0, 2.0, 1.0],
        [1.0, 2.0, nan, 1.0, 2.0, 2.0, 0.0, 2.0, 2.0, 2.0],
    ]
    assert_array_equal(G, arr)

    write_plink1_bin(G, tmp_path / "out.bed")

    with open(bed, "rb") as f:
        bed_desired = md5(f.read()).hexdigest()

    with open(tmp_path / "out.bed", "rb") as f:
        bed_content = md5(f.read()).hexdigest()

    assert bed_content == bed_desired

    bim_desired = "88475a9ea7a52e056716e612f44ccb62"
    with open(tmp_path / "out.bim", "rb") as f:
        bim_content = md5(f.read()).hexdigest()
    assert bim_content == bim_desired

    fam_desired = "2df7b9a70ab70e95f8b1c774b9022404"
    with open(tmp_path / "out.fam", "rb") as f:
        fam_content = md5(f.read()).hexdigest()
    assert fam_content == fam_desired


def test_write_plink1_bin_filename(tmp_path: Path):

    datafiles = join(dirname(realpath(__file__)), "data_files")
    file_prefix = join(datafiles, "data")
    bim = file_prefix + ".bim"
    bed = file_prefix + ".bed"
    fam = file_prefix + ".fam"

    G = read_plink1_bin(bed, bim, fam, verbose=False)
    bedfp = tmp_path / "out.bed"
    bimfp = tmp_path / "out.bim"
    famfp = tmp_path / "out.fam"
    write_plink1_bin(G, bedfp, bimfp, famfp)

    with open(bed, "rb") as f:
        bed_desired = md5(f.read()).hexdigest()

    with open(tmp_path / "out.bed", "rb") as f:
        bed_content = md5(f.read()).hexdigest()

    assert bed_content == bed_desired

    bim_desired = "88475a9ea7a52e056716e612f44ccb62"
    with open(tmp_path / "out.bim", "rb") as f:
        bim_content = md5(f.read()).hexdigest()
    assert bim_content == bim_desired

    fam_desired = "2df7b9a70ab70e95f8b1c774b9022404"
    with open(tmp_path / "out.fam", "rb") as f:
        fam_content = md5(f.read()).hexdigest()
    assert fam_content == fam_desired


def test_write_plink1_bin_transpose(tmp_path: Path):

    datafiles = join(dirname(realpath(__file__)), "data_files")
    file_prefix = join(datafiles, "data")
    bim = file_prefix + ".bim"
    bed = file_prefix + ".bed"
    fam = file_prefix + ".fam"

    G = read_plink1_bin(bed, bim, fam, verbose=False)
    write_plink1_bin(G.T, tmp_path / "out.bed")

    with open(bed, "rb") as f:
        bed_desired = md5(f.read()).hexdigest()

    with open(tmp_path / "out.bed", "rb") as f:
        bed_content = md5(f.read()).hexdigest()

    assert bed_content == bed_desired

    bim_desired = "88475a9ea7a52e056716e612f44ccb62"
    with open(tmp_path / "out.bim", "rb") as f:
        bim_content = md5(f.read()).hexdigest()
    assert bim_content == bim_desired

    fam_desired = "2df7b9a70ab70e95f8b1c774b9022404"
    with open(tmp_path / "out.fam", "rb") as f:
        fam_content = md5(f.read()).hexdigest()
    assert fam_content == fam_desired


def test_write_plink1_bin_sample_major(tmp_path: Path):

    datafiles = join(dirname(realpath(__file__)), "data_files")
    file_prefix = join(datafiles, "data")
    bim = file_prefix + ".bim"
    bed = file_prefix + ".bed"
    fam = file_prefix + ".fam"

    G = read_plink1_bin(bed, bim, fam, verbose=False)
    write_plink1_bin(G, tmp_path / "tmp.bed", major="sample", verbose=False)

    G = read_plink1_bin(str(tmp_path / "tmp.bed"), verbose=False)
    write_plink1_bin(G, tmp_path / "out.bed", verbose=False)

    with open(bed, "rb") as f:
        bed_desired = md5(f.read()).hexdigest()

    with open(tmp_path / "out.bed", "rb") as f:
        bed_content = md5(f.read()).hexdigest()

    assert bed_content == bed_desired

    bim_desired = "88475a9ea7a52e056716e612f44ccb62"
    with open(tmp_path / "out.bim", "rb") as f:
        bim_content = md5(f.read()).hexdigest()
    assert bim_content == bim_desired

    fam_desired = "2df7b9a70ab70e95f8b1c774b9022404"
    with open(tmp_path / "out.fam", "rb") as f:
        fam_content = md5(f.read()).hexdigest()
    assert fam_content == fam_desired


def test_write_plink1_bin_empty_metadata(tmp_path: Path):

    datafiles = join(dirname(realpath(__file__)), "data_files")
    file_prefix = join(datafiles, "data")
    bim = file_prefix + ".bim"
    bed = file_prefix + ".bed"
    fam = file_prefix + ".fam"

    G = read_plink1_bin(bed, bim, fam, verbose=False)
    names = list(set(G.coords.keys()) - set(["sample", "variant"]))
    G = G.drop_vars(names)
    write_plink1_bin(G, tmp_path / "out.bed")

    with open(bed, "rb") as f:
        bed_desired = md5(f.read()).hexdigest()

    with open(tmp_path / "out.bed", "rb") as f:
        bed_content = md5(f.read()).hexdigest()

    assert bed_content == bed_desired

    bim_desired = "3dd5c109ba236d8b770d9c29dbc23c14"
    with open(tmp_path / "out.bim", "rb") as f:
        bim_content = md5(f.read()).hexdigest()
    assert bim_content == bim_desired

    fam_desired = "37dcb1a777a2a99ca3484b82cfb9c33b"
    with open(tmp_path / "out.fam", "rb") as f:
        fam_content = md5(f.read()).hexdigest()
    assert fam_content == fam_desired



def pytest_configure(*_):
    setup_tests_baseline()


def pytest_runtest_setup():
    setup_tests_baseline()


def setup_tests_baseline():
    _compatibility()
    import doctest

    import numpy as np

    _pandas_format()
    doctest.ELLIPSIS_MARKER = "-ignore-"
    np.set_printoptions(precision=2, floatmode="fixed")


def _pandas_format():
    import pandas as pd

    pd.set_option("display.width", 88)
    pd.set_option("display.max_columns", 79)
    pd.set_option("display.max_rows", 60)
    pd.set_option("display.large_repr", "truncate")
    pd.set_option("display.float_format", "{:8.2f}".format)


def _compatibility():
    import warnings

    warnings.filterwarnings("ignore", message="numpy.dtype size changed")
    warnings.filterwarnings("ignore", message="numpy.ufunc size changed")



from ._filetype import file_type
from ._util import last_replace


def read_grm(filepath, id_filepath=None, n_snps_filepath=None):
    """
    Read GCTA realized relationship matrix files.

    A GRM file set consists of two or three files: (i) one containing the covariance
    matrix; (ii) one contaning sample IDs; and (iii) possibly another one containing
    the number of non-missing SNPs.

    It supports plain text, binary, and compressed files. The usual file extensions for
    those types are `.grm`, `grm.bin`, and `.grm.gz`, respectively.

    Example
    -------
    .. doctest::

        >>> from os.path import join
        >>> from pandas_plink import read_grm
        >>> from pandas_plink import get_data_folder
        >>> filepath = join(get_data_folder(), "grm-list", "plink2.grm")
        >>> id_filepath = join(get_data_folder(), "grm-list", "plink2.grm.id")
        >>> (K, n_snps) = read_grm(filepath, id_filepath)
        >>> print(K)
        <xarray.DataArray (sample_0: 10, sample_1: 10)>
        array([[ 0.89,  0.23, -0.19, -0.01, -0.14,  0.29,  0.27, -0.23, -0.10,
                -0.21],
               [ 0.23,  1.08, -0.45,  0.19, -0.19,  0.17,  0.41, -0.01, -0.13,
                -0.13],
               [-0.19, -0.45,  1.18, -0.04, -0.15, -0.20, -0.31, -0.04,  0.30,
                -0.01],
               [-0.01,  0.19, -0.04,  0.90, -0.07,  0.01,  0.06, -0.19, -0.09,
                 0.17],
               [-0.14, -0.19, -0.15, -0.07,  1.18,  0.09, -0.03,  0.10,  0.22,
                 0.17],
               [ 0.29,  0.17, -0.20,  0.01,  0.09,  0.96,  0.07, -0.04, -0.09,
                -0.23],
               [ 0.27,  0.41, -0.31,  0.06, -0.03,  0.07,  0.71, -0.10, -0.09,
                -0.06],
               [-0.23, -0.01, -0.04, -0.19,  0.10, -0.04, -0.10,  1.42, -0.30,
                -0.07],
               [-0.10, -0.13,  0.30, -0.09,  0.22, -0.09, -0.09, -0.30,  0.91,
                -0.02],
               [-0.21, -0.13, -0.01,  0.17,  0.17, -0.23, -0.06, -0.07, -0.02,
                 0.91]])
        Coordinates:
          * sample_0  (sample_0) object 'HG00419' 'HG00650' ... 'NA20508' 'NA20753'
          * sample_1  (sample_1) object 'HG00419' 'HG00650' ... 'NA20508' 'NA20753'
            fid       (sample_1) object 'HG00419' 'HG00650' ... 'NA20508' 'NA20753'
            iid       (sample_1) object 'HG00419' 'HG00650' ... 'NA20508' 'NA20753'
        >>> print(n_snps)
        [50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50
         50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50
         50 50 50 50 50 50 50]
        >>> filepath = join(get_data_folder(), "grm-bin", "plink.grm.bin")
        >>> id_filepath = join(get_data_folder(), "grm-bin", "plink.grm.id")
        >>> n_snps_filepath = join(get_data_folder(), "grm-bin", "plink.grm.N.bin")
        >>> (K, n_snps) = read_grm(filepath, id_filepath, n_snps_filepath)
        >>> print(K)
        <xarray.DataArray (sample_0: 10, sample_1: 10)>
        array([[ 0.79,  0.10, -0.19, -0.07, -0.27,  0.15,  0.20, -0.27, -0.18,
                -0.26],
               [ 0.10,  1.07, -0.45,  0.04, -0.34, -0.01,  0.23, -0.15, -0.25,
                -0.25],
               [-0.19, -0.45,  1.39, -0.07, -0.24, -0.23, -0.35, -0.09,  0.28,
                -0.05],
               [-0.07,  0.04, -0.07,  0.82, -0.16, -0.13, -0.05, -0.30, -0.17,
                 0.08],
               [-0.27, -0.34, -0.24, -0.16,  1.08, -0.10, -0.12, -0.03,  0.11,
                 0.06],
               [ 0.15, -0.01, -0.23, -0.13, -0.10,  0.94, -0.05, -0.11, -0.16,
                -0.31],
               [ 0.20,  0.23, -0.35, -0.05, -0.12, -0.05,  0.59, -0.18, -0.14,
                -0.13],
               [-0.27, -0.15, -0.09, -0.30, -0.03, -0.11, -0.18,  1.49, -0.32,
                -0.05],
               [-0.18, -0.25,  0.28, -0.17,  0.11, -0.16, -0.14, -0.32,  0.89,
                -0.06],
               [-0.26, -0.25, -0.05,  0.08,  0.06, -0.31, -0.13, -0.05, -0.06,
                 0.95]])
        Coordinates:
          * sample_0  (sample_0) object 'HG00419' 'HG00650' ... 'NA20508' 'NA20753'
          * sample_1  (sample_1) object 'HG00419' 'HG00650' ... 'NA20508' 'NA20753'
            fid       (sample_1) object 'HG00419' 'HG00650' ... 'NA20508' 'NA20753'
            iid       (sample_1) object 'HG00419' 'HG00650' ... 'NA20508' 'NA20753'
        >>> print(n_snps)
        [50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50
         50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50
         50 50 50 50 50 50 50]

    Parameters
    ----------
    filepath : str
        Path to the matrix file.
    id_filepath : str, optional
        Path to the file containing family and individual IDs. It defaults to ``None``,
        in which case it will try to be inferred.
    n_snps_filepath : str, optional
        Path to the file containing the number of non-missing SNPs. It defaults to
        ``None``, in which case it will try to be inferred.

    Returns
    -------
    grm : :class:`xarray.DataArray`
        Realized relationship matrix.
    n_snps : :class:`numpy.ndarray`
        Number of non-missing SNPs.
    """
    if file_type(filepath) == "bin":
        return _read_gcta_grm_bin(filepath, id_filepath, n_snps_filepath)
    return _read_gcta_grm(filepath, id_filepath)


def _read_gcta_grm(filepath, id_filepath):
    from numpy import asarray, int64, tril, zeros
    from pandas import read_csv
    from xarray import DataArray

    if filepath.endswith(".gz"):
        basename = filepath[:-3]
    else:
        basename = filepath

    if id_filepath is None:
        id_filepath = basename + ".id"

    df = read_csv(filepath, sep="\t", header=None)
    df_id = read_csv(id_filepath, sep="\t", header=None)
    n = int64(df.iloc[-1, 0])
    x = asarray(df.iloc[:, 0] - 1, int64)
    y = asarray(df.iloc[:, 1] - 1, int64)

    K = zeros((n, n))
    K[x, y] = df.iloc[:, 3]
    K = K + tril(K, -1).T
    coords = (df_id.iloc[:, 1], df_id.iloc[:, 1])
    K = DataArray(K, dims=["sample_0", "sample_1"], coords=coords)
    K = K.assign_coords(**{"fid": ("sample_0", df_id.iloc[:, 0])})
    K = K.assign_coords(**{"fid": ("sample_1", df_id.iloc[:, 0])})

    K = K.assign_coords(**{"iid": ("sample_0", df_id.iloc[:, 1])})
    K = K.assign_coords(**{"iid": ("sample_1", df_id.iloc[:, 1])})

    n_snps = asarray(df.iloc[:, 2], int64)

    return (K, n_snps)


def _read_gcta_grm_bin(filepath, id_filepath, n_snps_filepath):
    from numpy import (
        asarray,
        float32,
        float64,
        fromfile,
        int64,
        tril,
        tril_indices_from,
        zeros,
    )
    from pandas import read_csv
    from xarray import DataArray

    if id_filepath is None:
        id_filepath = last_replace(filepath, ".bin", ".id")

    if n_snps_filepath is None:
        n_snps_filepath = last_replace(filepath, ".bin", ".N.bin")

    df_id = read_csv(id_filepath, sep="\t", header=None)
    n = df_id.shape[0]
    k = asarray(fromfile(filepath, dtype=float32), float64)
    n_snps = asarray(fromfile(n_snps_filepath, dtype=float32), int64)

    K = zeros((n, n))
    K[tril_indices_from(K)] = k
    K = K + tril(K, -1).T
    coords = (df_id.iloc[:, 1], df_id.iloc[:, 1])
    K = DataArray(K, dims=["sample_0", "sample_1"], coords=coords)

    K = K.assign_coords(**{"fid": ("sample_0", df_id.iloc[:, 0])})
    K = K.assign_coords(**{"fid": ("sample_1", df_id.iloc[:, 0])})

    K = K.assign_coords(**{"iid": ("sample_0", df_id.iloc[:, 1])})
    K = K.assign_coords(**{"iid": ("sample_1", df_id.iloc[:, 1])})

    return (K, n_snps)



from dataclasses import dataclass
from typing import Optional

__all__ = ["Chunk"]


@dataclass
class Chunk:
    """
    Chunk specification.

    It is effectively a contiguous submatrix of the dosage matrix.

    Parameters
    ----------
    nsamples
        Number of samples in a single chunk, thresholded by the total
        number of samples. Set to ``None`` to include all samples.
        Default to ``1024``.
    nvariants
        Number of variants in a single chunk, thresholded by the total
        number of variants. Set to ``None`` to include all variants.
        Default to ``1024``.

    Note
    ----
    Small chunks might increase computation time while large chunks
    might increase IO usage. If you have a small data set, try setting
    both ``nsamples`` and ``nvariants`` to ``None``. If the data set
    is too large but your application will use every sample, try to set
    ``nsamples=None`` and choose a small value for ``nvariants``.

    Examples
    --------
    .. doctest::

        >>> from pandas_plink import Chunk
        >>>
        >>> Chunk()
        Chunk(nsamples=1024, nvariants=1024)
        >>> Chunk(nsamples=None)
        Chunk(nsamples=None, nvariants=1024)
    """

    nsamples: Optional[int] = 1024
    nvariants: Optional[int] = 1024



from os.path import dirname, join, realpath

from numpy.testing import assert_almost_equal, assert_equal

from pandas_plink import read_grm


def test_read_grm():
    datafiles = join(dirname(realpath(__file__)), "data_files")
    filepath = join(datafiles, "grm-list", "plink2.grm")
    (K, n_snps) = read_grm(filepath)

    assert_almost_equal([K.data[0, 0], K.data[-2, 5]], [0.885782, -0.0914917])
    assert_equal(K.sample_0[4].data, "NA12489")
    assert_equal(n_snps[3], 50)

    filepath = join(datafiles, "grm-list", "plink2.grm")
    filepath_id = join(datafiles, "grm-list", "plink2.grm.id")
    (K, n_snps) = read_grm(filepath, filepath_id)

    assert_almost_equal([K.data[0, 0], K.data[-2, 5]], [0.885782, -0.0914917])
    assert_equal(K.sample_0[4].data, "NA12489")
    assert_equal(n_snps[3], 50)


def test_read_grm_gz():
    datafiles = join(dirname(realpath(__file__)), "data_files")

    filepath = join(datafiles, "grm-gz", "plink.grm.gz")
    (K, n_snps) = read_grm(filepath)

    assert_almost_equal([K.data[0, 0], K.data[-2, 5]], [0.78974109888, -0.15841817856])
    assert_equal(K.sample_0[4].data, "NA12489")
    assert_equal(n_snps[3], 50)

    filepath = join(datafiles, "grm-gz", "plink.grm.gz")
    filepath_id = join(datafiles, "grm-gz", "plink.grm.id")
    (K, n_snps) = read_grm(filepath, filepath_id)

    assert_almost_equal([K.data[0, 0], K.data[-2, 5]], [0.78974109888, -0.15841817856])
    assert_equal(K.sample_0[4].data, "NA12489")
    assert_equal(n_snps[3], 50)


def test_read_grm_bin():
    datafiles = join(dirname(realpath(__file__)), "data_files")

    filepath = join(datafiles, "grm-bin", "plink.grm.bin")
    (K, n_snps) = read_grm(filepath)

    assert_almost_equal([K.data[0, 0], K.data[-2, 5]], [0.78974109888, -0.15841817856])
    assert_equal(K.sample_0[4].data, "NA12489")
    assert_equal(n_snps[3], 50)

    filepath = join(datafiles, "grm-bin", "plink.grm.bin")
    filepath_id = join(datafiles, "grm-bin", "plink.grm.id")
    (K, n_snps) = read_grm(filepath, filepath_id)

    assert_almost_equal([K.data[0, 0], K.data[-2, 5]], [0.78974109888, -0.15841817856])
    assert_equal(K.sample_0[4].data, "NA12489")
    assert_equal(n_snps[3], 50)

    filepath = join(datafiles, "grm-bin", "plink.grm.bin")
    filepath_id = join(datafiles, "grm-bin", "plink.grm.id")
    filepath_n_snps = join(datafiles, "grm-bin", "plink.grm.N.bin")
    (K, n_snps) = read_grm(filepath, filepath_id, filepath_n_snps)

    assert_almost_equal([K.data[0, 0], K.data[-2, 5]], [0.78974109888, -0.15841817856])
    assert_equal(K.sample_0[4].data, "NA12489")
    assert_equal(n_snps[3], 50)



def file_type(filepath):

    if _is_gzip(filepath):
        file_type = "gzip"
    elif _is_zstd(filepath):
        file_type = "zstd"
    elif _is_binary_file(filepath):
        file_type = "bin"
    else:
        file_type = "txt"

    return file_type


def _is_zstd(filepath):
    with open(filepath, "rb") as f:
        hdr = f.read(4)
        if int.from_bytes(hdr, byteorder="little") == 4247762216:
            return True
    return False


def _is_gzip(filepath):
    with open(filepath, "rb") as f:
        hdr = f.read(2)
        if int.from_bytes(hdr, byteorder="little") == 35615:
            return True
    return False


def _is_binary_file(filepath):
    with open(filepath, "rb") as f:
        bytes = f.read(1024)
    textchars = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7F})
    return bool(bytes.translate(None, textchars))
